// UBC CPSC 314 (2015W2) -- P1 HAVE FUN!!! :)

//GLOBAL VARS
// set before execution 
var debug = 0; // 1 = wireframe, 0 = normal rendering.
var torso_length = 8;
var head_size = 2;
var tail_segment_size = 0.5;
var front_paw_length = 3;
var front_paw_width = 3;
var back_paw_length = 2;
var back_paw_width = 2;
var front_claw_size = 1.5;
var back_claw_size = 1;
var tentacle_size = 1;
// to be changed by application
var useSmooth = 0; // 1 = smooth animation, 0 = jumpcut
// ASSIGNMENT-SPECIFIC API EXTENSION
THREE.Object3D.prototype.setMatrix = function(a) {
  this.matrix=a;
  this.matrix.decompose(this.position,this.quaternion,this.scale);
}

// SETUP RENDERER & SCENE
var canvas = document.getElementById('canvas');
var scene = new THREE.Scene();
var renderer = new THREE.WebGLRenderer();
renderer.setClearColor(0xFFFFFF); // white background colour
canvas.appendChild(renderer.domElement);

// SETUP CAMERA
var camera = new THREE.PerspectiveCamera(30,1,0.1,1000); // view angle, aspect ratio, near, far
camera.position.set(45,20,40);
camera.lookAt(scene.position);
scene.add(camera);

// SETUP ORBIT CONTROLS OF THE CAMERA
var controls = new THREE.OrbitControls(camera);

// ADAPT TO WINDOW RESIZE
function resize() {
  renderer.setSize(window.innerWidth,window.innerHeight);
  camera.aspect = window.innerWidth/window.innerHeight;
  camera.updateProjectionMatrix();
}

// EVENT LISTENER RESIZE
window.addEventListener('resize',resize);
resize();

//SCROLLBAR FUNCTION DISABLE
window.onscroll = function () {
     window.scrollTo(0,0);
   }

// SETUP HELPER GRID
// Note: Press Z to show/hide
var gridGeometry = new THREE.Geometry();
var i;
for(i=-50;i<51;i+=2) {
    gridGeometry.vertices.push( new THREE.Vector3(i,0,-50));
    gridGeometry.vertices.push( new THREE.Vector3(i,0,50));
    gridGeometry.vertices.push( new THREE.Vector3(-50,0,i));
    gridGeometry.vertices.push( new THREE.Vector3(50,0,i));
}

var gridMaterial = new THREE.LineBasicMaterial({color:0xBBBBBB});
var grid = new THREE.Line(gridGeometry,gridMaterial,THREE.LinePieces);

/////////////////////////////////
//   YOUR WORK STARTS BELOW    //
/////////////////////////////////

// MATERIALS
// Note: Feel free to be creative with this! 
var normalMaterial;
if (debug) {
  normalMaterial = new THREE.MeshBasicMaterial({
    color: 0x0000ff,
    wireframe: true
  });
} else {
  normalMaterial = new THREE.MeshNormalMaterial();
}

// function drawCube()
// Draws a unit cube centered about the origin.
// Note: You will be using this for all of your geometry
function makeCube() {
  var unitCube = new THREE.BoxGeometry(1,1,1);
  return unitCube;
}

// GEOMETRY
var torsoGeometry = makeCube();
var non_uniform_scale = new THREE.Matrix4().set(5,0,0,0, 0,5,0,0, 0,0,torso_length,0, 0,0,0,1);
torsoGeometry.applyMatrix(non_uniform_scale);

// TO-DO: SPECIFY THE REST OF YOUR STAR-NOSE MOLE'S GEOMETRY. 
// Note: You will be using transformation matrices to set the shape. 
// Note: You are not allowed to use the tools Three.js provides for 
//       rotation, translation and scaling.
// Note: The torso has been done for you (but feel free to modify it!)  
// Hint: Explicity declare new matrices using Matrix4().set     
var headGeometry = makeCube();
var head_uniform_scale = new THREE.Matrix4().set(head_size,0,0,0, 0,head_size,0,0, 0,0,head_size,0, 0,0,0,head_size);

var noseGeometry = makeCube();
var nose_scale = new THREE.Matrix4().set(2,0,0,0, 0,1,0,0, 0,0,0.5,0, 0,0,0,1);

var tailGeometry = makeCube();
function getTailScale(tailSize, tailWidth) {
  return new THREE.Matrix4().set(tailWidth,0,0,0, 0,tailWidth,0,0, 0,0,tailSize,0, 0,0,0,1);
}

var tail_scale = getTailScale(tail_segment_size, 0.5);

var frontPawGeometry = makeCube();
var front_paw_scale = new THREE.Matrix4().set(front_paw_width,0,0,0,0,1,0,0,0,0,front_paw_length,0,0,0,0,1);

var backPawGeometry = makeCube();
var back_paw_scale = new THREE.Matrix4().set(back_paw_width,0,0,0,0,1,0,0,0,0,back_paw_length, 0,0,0,0,1);

// MATRICES
var torsoMatrix = new THREE.Matrix4().set(1,0,0,0, 0,1,0,2.5, 0,0,1,0, 0,0,0,1);
headGeometry.applyMatrix(head_uniform_scale);
noseGeometry.applyMatrix(nose_scale);
tailGeometry.applyMatrix(tail_scale);
frontPawGeometry.applyMatrix(front_paw_scale);
backPawGeometry.applyMatrix(back_paw_scale);
// TO-DO: INITIALIZE THE REST OF YOUR MATRICES 
// Note: Use of parent attribute is not allowed.
// Hint: Keep hierarchies in mind!   
// Hint: Play around with the headTorsoMatrix values, what changes in the render? Why?         
var headMatrix = new THREE.Matrix4().set(1,0,0,0, 0,1,0,0, 0,0,1,5, 0,0,0,1);
var noseMatrix = new THREE.Matrix4().set(1,0,0,0, 0,1,0,0, 0,0,1,1.25, 0,0,0,1);

function getTailMatrix(translateAmount) {
  return new THREE.Matrix4().set(1,0,0,0, 0,1,0,0, 0,0,1,translateAmount, 0,0,0,1);
}

var tailMatrix = getTailMatrix(-(torso_length + tail_segment_size)/2);

var initialDownMagnitude = Math.PI/8;
var initialDownMatrix = new THREE.Matrix4().set(1, 0, 0, 0,
  0, Math.cos(initialDownMagnitude), -Math.sin(initialDownMagnitude), 0,
  0, Math.sin(initialDownMagnitude), Math.cos(initialDownMagnitude), 0,
  0, 0, 0, 1);

var frontPawMatrixLeft = new THREE.Matrix4().set(1, 0, 0, 2, 0, 1, 0, -2.5, 0, 0, 1, 4, 0, 0, 0, 1);
var frontPawMatrixRight = new THREE.Matrix4().set(1, 0, 0, -2, 0, 1, 0, -2.5, 0, 0, 1, 4, 0, 0, 0, 1);
var backPawMatrixLeft = new THREE.Matrix4().set(1,0,0,2, 0,1,0,-2.5, 0,0,1,-2.5, 0,0,0,1);
var backPawMatrixRight = new THREE.Matrix4().set(1,0,0,-2, 0,1,0,-2.5, 0,0,1,-2.5, 0,0,0,1);

frontPawMatrixLeft = new THREE.Matrix4().multiplyMatrices(frontPawMatrixLeft, initialDownMatrix);
frontPawMatrixRight = new THREE.Matrix4().multiplyMatrices(frontPawMatrixRight, initialDownMatrix);
backPawMatrixLeft = new THREE.Matrix4().multiplyMatrices(backPawMatrixLeft, initialDownMatrix);
backPawMatrixRight = new THREE.Matrix4().multiplyMatrices(backPawMatrixRight, initialDownMatrix);


//Each body part is represented by one of these objects
//relativeMatrix = position relative to parent body part
//staticRelativeMatrix = inactive (i.e. the position we return to the second time we hit a command) position relative to parent body part
//absoluteMatrix = position in the world
//staticAbsoluteMatrix = inactive position relative to world
//each Part object will be doubly linked with its parent
function Part(geometry, material, parent, matrix) {
  this.mesh = new THREE.Mesh(geometry, material);
  this.parentPart = parent;
  this.relativeMatrix = matrix;
  this.staticRelativeMatrix = matrix;
  if (parent == null) {
    this.absoluteMatrix = matrix;
    this.staticAbsoluteMatrix = matrix;
  } else {
    parent.childrenParts.push(this);
    this.absoluteMatrix = new THREE.Matrix4().multiplyMatrices(parent.absoluteMatrix, matrix);
    this.staticAbsoluteMatrix = this.absoluteMatrix;
  }
  this.childrenParts = [];
}

// CREATE BODY
//the torso will be the root part
var torso = new Part(torsoGeometry, normalMaterial, null, torsoMatrix);
var head = new Part(headGeometry, normalMaterial, torso, headMatrix);
var nose = new Part(noseGeometry, normalMaterial, head, noseMatrix);

//make tail
var tail = new Part(tailGeometry, normalMaterial, torso, tailMatrix);

var currSegment = tail;
for (var i = 1; i < 10; i++) {
  var tailCube = makeCube();
  tailCube.applyMatrix(getTailScale(tail_segment_size, 1/(1+i)));
  var newSegment = new Part(tailCube, normalMaterial, currSegment, getTailMatrix(-tail_segment_size));
  currSegment = newSegment;
}
///finish making tail

var leftFrontPaw = makePaw(frontPawGeometry, normalMaterial, torso, frontPawMatrixLeft, front_paw_width, front_claw_size);
var rightFrontPaw = new makePaw(frontPawGeometry, normalMaterial, torso, frontPawMatrixRight, front_paw_width, front_claw_size);
var leftBackPaw = makePaw(backPawGeometry, normalMaterial, torso, backPawMatrixLeft, back_paw_width, back_claw_size);
var rightBackPaw = makePaw(backPawGeometry, normalMaterial, torso, backPawMatrixRight, back_paw_width, back_claw_size);


function makePaw(geometry, material, parentPart, pawMatrix, pawWidth, clawSize) {
  var paw = new Part(geometry, material, parentPart, pawMatrix);
  clawSize = clawSize/2;
  var numClaws = 5;
  var unitDiff = pawWidth/numClaws;
  
  var clawGeometry = makeCube();
  var claw_scale = new THREE.Matrix4().set(0.25,0,0,0,0,0.75,0,0,0,0,clawSize,0,0,0,0,1);
  clawGeometry.applyMatrix(claw_scale);
  var forwardClawMatrix = new THREE.Matrix4().set(1,0,0,0,
                                                  0,0.5,0,0,
                                                  0,0,1,clawSize*2/3,
                                                  0,0,0,1);

  for (var i = -2; i < -2 + numClaws; i++) {
    var clawMatrix = new THREE.Matrix4().set(1,0,0,i*unitDiff, 0,1,0,0, 0,0,1,pawWidth*0.5, 0,0,0,1);
    var claw = new Part(clawGeometry, normalMaterial, paw, clawMatrix);
    var claw2 = new Part(clawGeometry, normalMaterial, claw, forwardClawMatrix);
    var claw3 = new Part(clawGeometry, normalMaterial, claw2, forwardClawMatrix);
  }

  return paw;
}

var tentacle_scale = new THREE.Matrix4().set(0.15,0,0,0, 0,0.15,0,0, 0,0,1,0, 0,0,0,1);
var tentacleGeometry = makeCube();
tentacleGeometry.applyMatrix(tentacle_scale);

var tentacleForwardMatrix = new THREE.Matrix4().set(1, 0, 0, 0,
  0, 1, 0, 0,
  0, 0, 1, tentacle_size / 2,
  0, 0, 0, 1);
var tentacleBackMatrix = new THREE.Matrix4().set(1, 0, 0, 0,
  0, 1, 0, 0,
  0, 0, 1, (-1) * tentacle_size / 2,
  0, 0, 0, 1);

function makeTentacles(leftRight) {
  //Nose tentacles rhs
  var posY = -0.5;
  var fan = Math.PI/6;
  var angle = 0;

  //This loop is for the main 18 tentacles on the side of the nose
  for (var i = 0; i < 9; i++) {
    angle += i*leftRight*(Math.PI/(36));

    var tentacleTranslate = new THREE.Matrix4().set(1, 0, 0, leftRight*(0.5 + 0.5 * Math.sin(0.4 * i)),
      0, 1, 0, posY,
      0, 0, 1, 0.5,
      0, 0, 0, 1);
    posY += 0.13;

   
    // This makes it into chainsaw mode
    // var tentacleRotate = new THREE.Matrix4().set(1,0,0,0, 0,Math.cos(i), -Math.sin(i),0, 0,Math.sin(i), Math.cos(i),0, 0,0,0,1);
    var tentacleRotate = new THREE.Matrix4().set(Math.cos(angle), -Math.sin(angle), 0, 0,
      Math.sin(angle), Math.cos(angle), 0, 0,
      0, 0, 1, 0,
      0, 0, 0, 1);

    var tentacleFan = new THREE.Matrix4().set(1, 0, 0, 0,
      0, Math.cos(fan), -Math.sin(fan), 0,
      0, Math.sin(fan), Math.cos(fan), 0,
      0, 0, 0, 1);
    tentacleTranslate = new THREE.Matrix4().multiplyMatrices(tentacleTranslate, tentacleRotate);

    tentacleTranslate = new THREE.Matrix4().multiplyMatrices(tentacleTranslate, tentacleBackMatrix);
    tentacleTranslate = new THREE.Matrix4().multiplyMatrices(tentacleTranslate, tentacleFan);
    tentacleTranslate = new THREE.Matrix4().multiplyMatrices(tentacleTranslate, tentacleForwardMatrix);
    var tentacle = new Part(tentacleGeometry, normalMaterial, nose, tentacleTranslate);
  }

  var topShortTentacleMatrix = getShortTentacleMatrix(0.2*leftRight, 0.5);

  var shortTentacleRotAmount = Math.PI;
  var rotate180Matrix = new THREE.Matrix4().set(Math.cos(shortTentacleRotAmount), -Math.sin(shortTentacleRotAmount), 0, 0,
      Math.sin(shortTentacleRotAmount), Math.cos(shortTentacleRotAmount), 0, 0,
      0, 0, 1, 0,
      0, 0, 0, 1);
  topShortTentacleMatrix = new THREE.Matrix4().multiplyMatrices(topShortTentacleMatrix, rotate180Matrix);

  var bottomShortTentacleMatrix = getShortTentacleMatrix(0.2*leftRight, -0.5);
    
  var tentacleShortTop = new Part(tentacleGeometry, normalMaterial, nose, topShortTentacleMatrix);
  var tentacleShortBottom = new Part(tentacleGeometry, normalMaterial, nose, bottomShortTentacleMatrix);

}

function getShortTentacleMatrix(shortX, shortY) {
  return new THREE.Matrix4().set(1, 0, 0, shortX,
    0, 1, 0, shortY,
    0, 0, 1, 0.5,
    0, 0, 0, 1);
}

makeTentacles(1);
makeTentacles(-1);

//Recursively add the entire hierarchy to the scene
function addHierarchyToScene(part) {
  part.mesh.setMatrix(part.absoluteMatrix); //root
  
  scene.add(part.mesh);

  for (var i = 0; i < part.childrenParts.length; i++) {
    addHierarchyToScene(part.childrenParts[i]);
  }
}

addHierarchyToScene(torso);
// TO-DO: PUT TOGETHER THE REST OF YOUR STAR-NOSED MOLE AND ADD TO THE SCENE!
// Hint: Hint: Add one piece of geometry at a time, then implement the motion for that part. 
//             Then you can make sure your hierarchy still works properly after each step.



// APPLY DIFFERENT JUMP CUTS/ANIMATIONS TO DIFFERNET KEYS
// Note: The start of "U" animation has been done for you, you must implement the hiearchy and jumpcut.
// Hint: There are other ways to manipulate and grab clock values!!
// Hint: Check THREE.js clock documenation for ideas.
// Hint: It may help to start with a jumpcut and implement the animation after.
// Hint: Where is updateBody() called?
var clock = new THREE.Clock(true);
var animations = {};

// Creates an animation. isReturning is a boolean that describes if the animation is returning a body part to its original position.
function Animation(p_start,p_end,t_length,isReturning) {
  this.p0 = p_start;
  this.p1 = p_end;
  this.time_length = t_length;
  this.time_start = clock.getElapsedTime();
  this.time_end = this.time_start + this.time_length;
  this.animate = true; // flag for animation
  this.isReturning = isReturning;
}

// function init_animation()
// Adds an Animation to the animations array, which is updated in doAnimations(). isReturning is a boolean that describes if the animation is returning a body part to its original position.
function init_animation(p_start,p_end,t_length,key,isReturning) {
  if (key == "S" && animations["S"] != null && animations["S"].swimState != 2) {
  	var swimState = animations["S"].swimState;
  	animations[key] = new Animation(p_start, p_end, t_length, isReturning);
  	animations[key].swimState = swimState;
  } else {
		animations[key] = new Animation(p_start, p_end, t_length, isReturning);
		if (key == "S") {
			animations[key].swimState = 0;
		}
  }
}

// update a part's absolute matrices based on its parent's absolute matrix, then repeat for the part's children
// this allows child body parts to move along with parent body parts
function updateChildrenAbsoluteMatrices(part) {
  part.absoluteMatrix = new THREE.Matrix4().multiplyMatrices(part.parentPart.absoluteMatrix, part.relativeMatrix);
  part.staticAbsoluteMatrix = new THREE.Matrix4().multiplyMatrices(part.parentPart.absoluteMatrix, part.staticRelativeMatrix);
  part.mesh.setMatrix(part.absoluteMatrix);
  for (var i = 0; i < part.childrenParts.length; i++) {
    updateChildrenAbsoluteMatrices(part.childrenParts[i]);
  }
}

//updates "animate" and "p"
function setFrameAndAnimation(animation) {
  if (useSmooth) {
    var time = clock.getElapsedTime(); // t seconds passed since the clock started.

    if (time > animation.time_end){
      animation.p = animation.p1;
      animation.animate = false;
      return 1; //needs to break
    }

    animation.p = (animation.p1 - animation.p0)*((time-animation.time_start)/animation.time_length) + animation.p0; // current frame    
  } else {
    animation.animate = false;
    animation.p = animation.p1;
    //we want to stop 
  }
  return 0; //no break
}

// move a part away from its static position, and update its children according to its new position
function movePartAndChildren(part, incrementalTransformMatrix, backMatrix, forwardMatrix) {

  var aboutToRotate = new THREE.Matrix4().multiplyMatrices(part.staticAbsoluteMatrix, backMatrix);
  var rotated = new THREE.Matrix4().multiplyMatrices(aboutToRotate, incrementalTransformMatrix);
  var backToPosition = new THREE.Matrix4().multiplyMatrices(rotated, forwardMatrix);
  part.absoluteMatrix = backToPosition;

  part.mesh.setMatrix(part.absoluteMatrix);
  
  var relativeAboutToRotate = new THREE.Matrix4().multiplyMatrices(part.staticRelativeMatrix, backMatrix);
  var relativeRotated = new THREE.Matrix4().multiplyMatrices(relativeAboutToRotate, incrementalTransformMatrix);
  var relativeBackToPosition = new THREE.Matrix4().multiplyMatrices(relativeRotated, forwardMatrix);

  part.relativeMatrix = relativeBackToPosition;

  for (var i = 0; i < part.childrenParts.length; i++) {
    updateChildrenAbsoluteMatrices(part.childrenParts[i]);        
  }
}

// do another iteration of all the active animations
function doAnimations() {
  for (var key in animations) {
    // skip loop if the property is from prototype
    if (!animations.hasOwnProperty(key)) continue;

    var animation = animations[key];
    updateBody(animation, key);
  }
}

//pivot matrices
var headBackMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,(-1)*head_size/2,
                                         0,0,0,1);
var headForwardMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,head_size/2,
                                         0,0,0,1);
var identity = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,0,
                                         0,0,0,1);
var tailBackMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,tail_segment_size/2,
                                         0,0,0,1);
var tailForwardMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,(-1)*tail_segment_size/2,
                                         0,0,0,1);
var frontPawBackMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,(-1)*front_paw_length/2,
                                         0,0,0,1);
var frontPawForwardMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,front_paw_length/2,
                                         0,0,0,1);
var backPawBackMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,(-1)*back_paw_length/2,
                                         0,0,0,1);
var backPawForwardMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,back_paw_length/2,
                                         0,0,0,1);
var clawBackMatrix = new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,(-1)*front_claw_size/2,
                                         0,0,0,1);
var clawForwardMatrix =  new THREE.Matrix4().set(1,0,0,0,
                                         0,1,0,0,
                                         0,0,1,front_claw_size/2,
                                         0,0,0,1);

function turnTail(tailTurnAmount) {
  var tailTurnMatrix = new THREE.Matrix4().set(Math.cos(tailTurnAmount), 0, Math.sin(tailTurnAmount), 0,
    0, 1, 0, 0, -Math.sin(tailTurnAmount), 0, Math.cos(tailTurnAmount), 0,
    0, 0, 0, 1);
  movePartAndChildren(tail, tailTurnMatrix, tailBackMatrix, tailForwardMatrix);
  curl(tail, tailTurnMatrix, tailBackMatrix, tailForwardMatrix);
}

function updateBody(animation, key) {
  switch(true)
  {
      case(key == "U" && animation.animate):
      
      if (setFrameAndAnimation(animation))
        break;
      
      var turnTorsoUp = new THREE.Matrix4().set(1,        0,         0,        0, 
                                              0, Math.cos(-(animation.p)),-Math.sin(-(animation.p)), 0, 
                                              0, Math.sin(-(animation.p)), Math.cos(-(animation.p)), 0,
                                              0,        0,         0,        1);
      
      movePartAndChildren(torso, turnTorsoUp, identity, identity);
      break;
      case (key == "E" && animation.animate):
      
      if (setFrameAndAnimation(animation))
        break;
      
      var turnTorsoDown = new THREE.Matrix4().set(1,        0,         0,        0, 
                                              0, Math.cos(animation.p),-Math.sin(animation.p), 0, 
                                              0, Math.sin(animation.p), Math.cos(animation.p), 0,
                                              0,        0,         0,        1);
      movePartAndChildren(torso, turnTorsoDown, identity, identity);
      // TO-DO: IMPLEMENT JUMPCUT/ANIMATION FOR EACH KEY!
      // Note: Remember spacebar sets jumpcut/animate!
      break;
      case (key == "H" && animation.animate):
      if (setFrameAndAnimation(animation))
        break;

      var turnHeadRight = new THREE.Matrix4().set(Math.cos(-(animation.p)), 0, Math.sin(-(animation.p)),           0,
                                            0,         1,             0,           0, 
                                            -Math.sin(-(animation.p)), 0,  Math.cos(-(animation.p)),         0,
                                              0,        0,            0,           1);
      
      movePartAndChildren(head, turnHeadRight, headBackMatrix, headForwardMatrix);

      break;
      case (key == "G" && animation.animate):
      if (setFrameAndAnimation(animation))
        break;

      var turnHeadLeft = new THREE.Matrix4().set(Math.cos(animation.p), 0, Math.sin(animation.p),           0,
                                            0,         1,             0,           0, 
                                            -Math.sin(animation.p), 0,  Math.cos(animation.p),         0,
                                              0,        0,            0,           1);
      movePartAndChildren(head, turnHeadLeft, headBackMatrix, headForwardMatrix);

      break;
      case (key == "N" && animation.animate):
      if (setFrameAndAnimation(animation))
        break;

      var fanAmount = animation.p;
      var tentacleFan = new THREE.Matrix4().set(1, 0, 0, 0,
      0, Math.cos(fanAmount), -Math.sin(fanAmount), 0,
      0, Math.sin(fanAmount), Math.cos(fanAmount), 0,
      0, 0, 0, 1);

      for (var i = 0; i < nose.childrenParts.length; i++) {
        movePartAndChildren(nose.childrenParts[i], tentacleFan, tentacleBackMatrix, tentacleForwardMatrix);
      }

      break;
      case (key == "V" && animation.animate):
      if (setFrameAndAnimation(animation))
        break;

      turnTail(-animation.p/5);

      break;
      case (key == "T" && animation.animate):
      if (setFrameAndAnimation(animation))
        break;
      turnTail(animation.p/5);
      break;
      case (key == "S" && animation.animate):
      if (setFrameAndAnimation(animation))
        break;
      if (animation.swimState == 0) {
	      var turnPawDown = new THREE.Matrix4().set(1,        0,         0,        0, 
	                                              0, Math.cos(animation.p),-Math.sin(animation.p), 0, 
	                                              0, Math.sin(animation.p), Math.cos(animation.p), 0,
	                                              0,        0,         0,        1);
	      movePartAndChildren(leftFrontPaw, turnPawDown, frontPawBackMatrix, frontPawForwardMatrix);
	      movePartAndChildren(rightBackPaw, turnPawDown, backPawBackMatrix, backPawForwardMatrix);
	      
        turnTail(-(animation.p/5));
	      var turnHeadRight = new THREE.Matrix4().set(Math.cos(-(animation.p)), 0, Math.sin(-(animation.p)),           0,
	                                            0,         1,             0,           0, 
	                                            -Math.sin(-(animation.p)), 0,  Math.cos(-(animation.p)),         0,
	                                              0,        0,            0,           1);
	      
	      movePartAndChildren(head, turnHeadRight, headBackMatrix, headForwardMatrix);
        var fanAmount = animation.p;
        var tentacleFan = new THREE.Matrix4().set(1, 0, 0, 0,
          0, Math.cos(fanAmount), -Math.sin(fanAmount), 0,
          0, Math.sin(fanAmount), Math.cos(fanAmount), 0,
          0, 0, 0, 1);

        for (var i = 0; i < nose.childrenParts.length; i++) {
          movePartAndChildren(nose.childrenParts[i], tentacleFan, tentacleBackMatrix, tentacleForwardMatrix);
        }

      } else if (animation.swimState == 1) {
      	var turnPawsBack = new THREE.Matrix4().set(1,        0,         0,        0, 
	                                              0, Math.cos(animation.p),-Math.sin(animation.p), 0, 
	                                              0, Math.sin(animation.p), Math.cos(animation.p), 0,
	                                              0,        0,         0,        1);
	      movePartAndChildren(leftFrontPaw, turnPawsBack, frontPawBackMatrix, frontPawForwardMatrix);
	      movePartAndChildren(rightBackPaw, turnPawsBack, backPawBackMatrix, backPawForwardMatrix);

	      var inverseP = animation.p0 - animation.p;
	      var otherPawsMove = new THREE.Matrix4().set(1,        0,         0,        0, 
	                                              0, Math.cos(inverseP),-Math.sin(inverseP), 0, 
	                                              0, Math.sin(inverseP), Math.cos(inverseP), 0,
	                                              0,        0,         0,        1);
	      movePartAndChildren(rightFrontPaw, otherPawsMove, frontPawBackMatrix, frontPawForwardMatrix);
	      movePartAndChildren(leftBackPaw, otherPawsMove, backPawBackMatrix, backPawForwardMatrix);

	      var moveHeadP = (animation.p0 - 2*animation.p);
        turnTail((animation.p0 - 2*animation.p)/5);

	      var turnHeadRight = new THREE.Matrix4().set(Math.cos(moveHeadP), 0, Math.sin(moveHeadP),           0,
	                                            0,         1,             0,           0, 
	                                            -Math.sin(moveHeadP), 0,  Math.cos(moveHeadP),         0,
	                                              0,        0,            0,           1);
	      
	      movePartAndChildren(head, turnHeadRight, headBackMatrix, headForwardMatrix); 	
      } else if (animation.swimState == 2) {
      	var otherPawsMove = new THREE.Matrix4().set(1,        0,         0,        0, 
	                                              0, Math.cos(animation.p),-Math.sin(animation.p), 0, 
	                                              0, Math.sin((animation.p)), Math.cos(animation.p), 0,
	                                              0,        0,         0,        1);
	      movePartAndChildren(rightFrontPaw, otherPawsMove, frontPawBackMatrix, frontPawForwardMatrix);
	      movePartAndChildren(leftBackPaw, otherPawsMove, backPawBackMatrix, backPawForwardMatrix);
	      
        turnTail(animation.p/5);
	      var turnHeadBack = new THREE.Matrix4().set(Math.cos(animation.p), 0, Math.sin(animation.p),           0,
	                                            0,         1,             0,           0, 
	                                            -Math.sin(animation.p), 0,  Math.cos(animation.p),         0,
	                                              0,        0,            0,           1);
	      
	      movePartAndChildren(head, turnHeadBack, headBackMatrix, headForwardMatrix);
        var fanAmount = animation.p;
        var tentacleFan = new THREE.Matrix4().set(1, 0, 0, 0,
          0, Math.cos(fanAmount), -Math.sin(fanAmount), 0,
          0, Math.sin(fanAmount), Math.cos(fanAmount), 0,
          0, 0, 0, 1);

        for (var i = 0; i < nose.childrenParts.length; i++) {
          movePartAndChildren(nose.childrenParts[i], tentacleFan, tentacleBackMatrix, tentacleForwardMatrix);
        }
      }

      break;
      case (key == "D" && animation.animate):
      if (setFrameAndAnimation(animation))
        break;
      var turnFrontPaws = new THREE.Matrix4().set(1,        0,         0,        0, 
                                              0, Math.cos((animation.p/2)),-Math.sin((animation.p/2)), 0, 
                                              0, Math.sin((animation.p/2)), Math.cos((animation.p/2)), 0,
                                              0,        0,         0,        1);
      movePartAndChildren(leftFrontPaw, turnFrontPaws, frontPawBackMatrix, frontPawForwardMatrix);
      movePartAndChildren(rightFrontPaw, turnFrontPaws, frontPawBackMatrix, frontPawForwardMatrix);
      curl(leftFrontPaw, turnFrontPaws, clawBackMatrix, clawForwardMatrix);
      curl(rightFrontPaw, turnFrontPaws, clawBackMatrix, clawForwardMatrix);

      break;

    default:
      break;
  }
}

function curl(part, rotMatrix, backMatrix, forwardMatrix) {
  for (var i = 0 ; i < part.childrenParts.length; i++) {
    movePartAndChildren(part.childrenParts[i], rotMatrix, backMatrix, forwardMatrix);
    curl(part.childrenParts[i], rotMatrix, backMatrix, forwardMatrix);
  }
}

// defines the toggle behaviour for body part movement shortcuts
function handleKey(p_start, p_end, t_length, key) {

	if (animations[key] != null && !animations[key].isReturning) {
    if (key == "S") {
    	if (animations[key].swimState == 0) {
    		init_animation(animations[key].p1,
				animations[key].p0,
				animations[key].time_length, key, false); //we don't return immediately	
    		animations[key].swimState = 1;
    	} else if (animations[key].swimState = 1){
    		init_animation(animations[key].p0,
				animations[key].p1,
				animations[key].time_length, key, true);
				animations[key].swimState = 2;
    	}
    } else {
			init_animation(animations[key].p1,
				animations[key].p0,
				animations[key].time_length, key, true);
    }
  } else {
    init_animation(p_start,p_end,t_length, key, false);
  }
}

// LISTEN TO KEYBOARD
// Hint: Pay careful attention to how the keys already specified work!
var keyboard = new THREEx.KeyboardState();
var grid_state = false;

keyboard.domElement.addEventListener('keydown',function(event){
  if (event.repeat)
    return;
  if(keyboard.eventMatches(event,"Z")){  // Z: Reveal/Hide helper grid
    grid_state = !grid_state;
    grid_state? scene.add(grid) : scene.remove(grid);
  } else if (keyboard.eventMatches(event,"0")) {    // 0: Set camera to neutral position, view reset
    camera.position.set(45,0,0);
    camera.lookAt(scene.position);
  } else if (keyboard.eventMatches(event,"U")) { 
    handleKey(0, Math.PI/4, 1, "U");    
  } else if (keyboard.eventMatches(event,"E")) {
    handleKey(0, Math.PI/4, 1, "E");
  } else if (keyboard.eventMatches(event,"H")) {
    handleKey(0, Math.PI/4, 1, "H");
  } else if (keyboard.eventMatches(event,"G")) {
    handleKey(0, Math.PI/4, 1, "G");
  } else if (keyboard.eventMatches(event,"T")) {
    handleKey(0, Math.PI/4, 1, "T");
  } else if (keyboard.eventMatches(event,"V")) {
    handleKey(0, Math.PI/4, 1, "V");
  } else if (keyboard.eventMatches(event,"N")) {
    handleKey(0, Math.PI/4, 1, "N");
  } else if (keyboard.eventMatches(event,"S")) {
    handleKey(0, Math.PI/4, 1, "S");
  } else if (keyboard.eventMatches(event,"D")) {
    handleKey(0, Math.PI/4, 1, "D");
  } else if (keyboard.eventMatches(event, " ")) {
    useSmooth = !useSmooth;
  }


  // TO-DO: BIND KEYS TO YOUR JUMP CUTS AND ANIMATIONS
  // Note: Remember spacebar sets jumpcut/animate! 
  // Hint: Look up "threex.keyboardstate by Jerome Tienne" for more info.
});

// SETUP UPDATE CALL-BACK
// Hint: It is useful to understand what is being updated here, the effect, and why.
function update() {
  doAnimations();

  requestAnimationFrame(update);
  renderer.render(scene,camera);
}

update();